"""
This module provides a bit of ontological information about the relations
that we support in ConceptNet. Every relation that ConceptNet supports
appears in this module.

Functions:
- is_negative_relation(rel: str) -> bool: Determine whether a relation is negative.
"""

from typing import Dict, TypeVar, Union

__all__ = [
    "ALL_RELATIONS",
    "SYMMETRIC_RELATIONS",
    "OPPOSITE_RELATIONS",
    "ENTAILED_RELATIONS",
    "is_negative_relation",
]


K = TypeVar("K")
V = TypeVar("V")


def _make_symmetric_dict(d: Dict[K, V]) -> Dict[Union[K, V], Union[V, K]]:
    """
    Given a dictionary, return a new dictionary that includes the reverse
    of every pair in the original dictionary. This is useful for relations
    that are symmetric, such as /r/SimilarTo.

    Parameters
    ----------
    d : Dict[K, V]
        A dictionary to mirror.

    Returns
    -------
    Dict[Union[K, V], Union[V, K]]
        A new dictionary that includes the reverse of every pair in the original dictionary.
    """
    newdict = dict(d)
    for key, val in d.items():
        newdict[val] = key
    return newdict


ALL_RELATIONS = [
    "/r/Antonym",
    "/r/AtLocation",
    "/r/CapableOf",
    "/r/Causes",
    "/r/CausesDesire",
    "/r/CreatedBy",
    "/r/DefinedAs",
    "/r/DerivedFrom",
    "/r/Desires",
    "/r/DistinctFrom",
    "/r/Entails",
    "/r/EtymologicallyDerivedFrom",
    "/r/EtymologicallyRelatedTo",
    "/r/ExternalURL",
    "/r/FormOf",
    "/r/HasA",
    "/r/HasContext",
    "/r/HasFirstSubevent",
    "/r/HasLastSubevent",
    "/r/HasPrerequisite",
    "/r/HasProperty",
    "/r/HasSubevent",
    "/r/InstanceOf",
    "/r/IsA",
    "/r/LocatedNear",
    "/r/MadeOf",
    "/r/MannerOf",
    "/r/MotivatedByGoal",
    "/r/NotCapableOf",
    "/r/NotDesires",
    "/r/NotHasProperty",
    "/r/NotUsedFor",
    "/r/ObstructedBy",
    "/r/PartOf",
    "/r/ReceivesAction",
    "/r/RelatedTo",
    "/r/SimilarTo",
    "/r/SymbolOf",
    "/r/Synonym",
    "/r/UsedFor",
    "/r/dbpedia/capital",
    "/r/dbpedia/field",
    "/r/dbpedia/genre",
    "/r/dbpedia/genus",
    "/r/dbpedia/influencedBy",
    "/r/dbpedia/knownFor",
    "/r/dbpedia/language",
    "/r/dbpedia/leader",
    "/r/dbpedia/occupation",
    "/r/dbpedia/product",
]

# These relations are symmetric: it doesn't matter which concept is their
# 'start' or 'end'.
SYMMETRIC_RELATIONS = {
    "/r/RelatedTo",
    "/r/SimilarTo",
    "/r/EtymologicallyRelatedTo",
    "/r/Synonym",
    "/r/Antonym",
    "/r/DistinctFrom",
}


# These relations are opposites of each other. Generally, a pair of terms
# having one relation implies that they don't have the opposite relation.
# You could consider these relations themselves to be related by the
# /r/Antonym relation.
OPPOSITE_RELATIONS = _make_symmetric_dict({
    "/r/NotDesires": "/r/Desires",
    "/r/NotUsedFor": "/r/UsedFor",
    "/r/NotCapableOf": "/r/CapableOf",
    "/r/NotHasProperty": "/r/HasProperty",
    "/r/Antonym": "/r/Synonym",
    "/r/ObstructedBy": "/r/HasPrerequisite",
})


# Most relations can be generalized into less specific relations. They don't
# form a strict tree in this way, but if you follow the chain of entailed
# relations, you'll usually get to either the very general /r/RelatedTo
# or /r/DistinctFrom.
#
# You could consider these relations themselves to be related by the
# /r/Entails relation.
#
# This mapping is not currently used in the ConceptNet code, but it could
# be used either in querying or in learning about relations.
ENTAILED_RELATIONS = {
    "/r/Antonym": "/r/DistinctFrom",
    "/r/Causes": "/r/RelatedTo",
    "/r/CausesDesire": "/r/RelatedTo",
    "/r/CapableOf": "/r/RelatedTo",
    "/r/CreatedBy": "/r/RelatedTo",
    "/r/DerivedFrom": "/r/RelatedTo",
    "/r/EtymologicallyRelatedTo": "/r/RelatedTo",
    "/r/Entails": "/r/RelatedTo",  # can we connect entailment and sub-events?
    "/r/HasContext": "/r/RelatedTo",
    "/r/HasProperty": "/r/RelatedTo",
    "/r/HasSubevent": "/r/RelatedTo",
    "/r/HasAgent": "/r/RelatedTo",
    "/r/HasTheme": "/r/RelatedTo",
    "/r/IsA": "/r/RelatedTo",
    "/r/LocatedNear": "/r/RelatedTo",
    "/r/MotivatedByGoal": "/r/RelatedTo",
    "/r/ReceivesAction": "/r/RelatedTo",
    "/r/SimilarTo": "/r/RelatedTo",
    "/r/SymbolOf": "/r/RelatedTo",
    "/r/UsedFor": "/r/RelatedTo",
    "/r/dbpedia": "/r/RelatedTo",
    "/r/FormOf": "/r/DerivedFrom",
    "/r/HasFirstSubevent": "/r/HasSubevent",
    "/r/HasLastSubevent": "/r/HasSubevent",
    "/r/HasPrerequisite": "/r/HasSubevent",
    "/r/MannerOf": "/r/Entails",
    "/r/DefinedAs": "/r/IsA",
    "/r/InstanceOf": "/r/IsA",
    "/r/AtLocation": "/r/LocatedNear",
    "/r/HasA": "/r/LocatedNear",
    "/r/PartOf": "/r/AtLocation",
    "/r/MadeOf": "/r/HasA",
    "/r/Synonym": "/r/SimilarTo",
}


def is_negative_relation(rel: str) -> bool:
    """
    Determine whether a relation is negative.

    Parameters
    ----------
    rel : str
        The relation to check, such as '/r/Antonym'.

    Returns
    -------
    bool
        True if the relation is negative, False otherwise.
    """
    return rel.startswith("/r/Not") or rel == "/r/Antonym" or rel == "/r/DistinctFrom"
